# Copyright: (c) 2022 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from krb5._context import Context

def cc_support_switch(
    context: Context,
    cache_type: bytes,
) -> bool:
    """Check whether the cache type supports switching.

    Checks whether the credential cache type specified supports switching the
    primary cache in its colleciton using :meth:`cc_switch`.

    Args:
        context: Krb5 context.
        cache_type: The credential cache type, like ``FILE``, ``DIR``, etc to
            check whether it supports switching or not.

    Returns:
        bool: The cache type supports switching.
    """



# Copyright: (c) 2021 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from krb5._context import Context

class Krb5Error(Exception):
    """Base Keberos Error class."""

    def __init__(
        self,
        context: Context,
        err_code: int,
    ) -> None: ...
    err_code: int  #: The Kerberos error code.



# Copyright: (c) 2021 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import typing

import k5test
import pytest

import krb5


class MockPrompt(krb5.Krb5Prompt):
    def __init__(self, responses: typing.List[bytes]) -> None:
        self.init_calls: typing.List[typing.Tuple[typing.Optional[bytes], typing.Optional[bytes], int]] = []
        self.prompt_calls: typing.List[typing.Tuple[bytes, bool]] = []
        self._responses = responses

    def init(
        self,
        name: typing.Optional[bytes],
        banner: typing.Optional[bytes],
        num_prompts: int,
    ) -> None:
        self.init_calls.append((name, banner, num_prompts))

    def prompt(self, msg: bytes, hidden: bool) -> bytes:
        self.prompt_calls.append((msg, hidden))
        return self._responses.pop(0)


def test_get_init_creds_keytab(realm: k5test.K5Realm) -> None:
    ctx = krb5.init_context()
    princ = krb5.parse_name_flags(ctx, realm.host_princ.encode())
    opt = krb5.get_init_creds_opt_alloc(ctx)
    kt = krb5.kt_default(ctx)

    creds = krb5.get_init_creds_keytab(ctx, princ, opt, kt)
    assert isinstance(creds, krb5.Creds)
    assert str(creds) == "Creds"

    assert creds.client.name == realm.host_princ.encode()
    assert creds.server.name == b"krbtgt/KRBTEST.COM@KRBTEST.COM"
    assert len(creds.keyblock.data) > 0
    assert str(creds.times).startswith("TicketTimes(authtime=")
    # creds.ticket_flags
    # creds.addresses
    assert len(creds.ticket) > 0
    assert creds.second_ticket == b""
    # creds.authdata


def test_get_init_creds_password(realm: k5test.K5Realm) -> None:
    ctx = krb5.init_context()
    princ = krb5.parse_name_flags(ctx, realm.user_princ.encode())
    opt = krb5.get_init_creds_opt_alloc(ctx)

    creds = krb5.get_init_creds_password(ctx, princ, opt, realm.password("user").encode())
    assert isinstance(creds, krb5.Creds)
    assert str(creds) == "Creds"


def test_get_init_creds_password_prompt(realm: k5test.K5Realm) -> None:
    ctx = krb5.init_context()
    princ = krb5.parse_name_flags(ctx, realm.user_princ.encode())
    opt = krb5.get_init_creds_opt_alloc(ctx)
    prompter = MockPrompt([realm.password("user").encode()])

    creds = krb5.get_init_creds_password(ctx, princ, opt, prompter=prompter)
    assert isinstance(creds, krb5.Creds)
    assert str(creds) == "Creds"

    assert len(prompter.init_calls) == 1
    assert prompter.init_calls[0] == (None, None, 1)
    assert len(prompter.prompt_calls) == 1

    expected = (
        f"Password for {realm.user_princ}" if realm.provider == "mit" else f"{realm.user_princ}'s Password: "
    ).encode()
    assert prompter.prompt_calls[0] == (expected, True)


def test_get_init_creds_password_prompt_failure(realm: k5test.K5Realm) -> None:
    ctx = krb5.init_context()
    princ = krb5.parse_name_flags(ctx, realm.user_princ.encode())
    opt = krb5.get_init_creds_opt_alloc(ctx)
    prompter = MockPrompt([])

    with pytest.raises(krb5.Krb5Error):
        krb5.get_init_creds_password(ctx, princ, opt, prompter=prompter)

    assert len(prompter.init_calls) == 1
    assert prompter.init_calls[0] == (None, None, 1)
    assert len(prompter.prompt_calls) == 1


def test_get_creds_keytab(realm: k5test.K5Realm) -> None:
    ctx = krb5.init_context()
    princ = krb5.parse_name_flags(ctx, realm.host_princ.encode())
    creds_ctx = krb5.init_creds_init(ctx, princ)

    assert isinstance(creds_ctx, krb5.InitCredsContext)
    assert str(creds_ctx) == "InitCredsContext"

    kt = krb5.kt_resolve(ctx, realm.keytab.encode())
    krb5.init_creds_set_keytab(ctx, creds_ctx, kt)
    krb5.init_creds_get(ctx, creds_ctx)

    creds = krb5.init_creds_get_creds(ctx, creds_ctx)
    assert isinstance(creds, krb5.Creds)
    assert str(creds) == "Creds"


def test_get_creds_keytab_wrong_principal(realm: k5test.K5Realm) -> None:
    ctx = krb5.init_context()
    princ = krb5.parse_name_flags(ctx, realm.user_princ.encode())
    creds_ctx = krb5.init_creds_init(ctx, princ)

    assert isinstance(creds_ctx, krb5.InitCredsContext)
    assert str(creds_ctx) == "InitCredsContext"

    kt = krb5.kt_resolve(ctx, realm.keytab.encode())

    expected = "no suitable keys for" if realm.provider == "mit" else "Failed to find"
    with pytest.raises(krb5.Krb5Error, match=expected):
        krb5.init_creds_set_keytab(ctx, creds_ctx, kt)


def test_init_creds_set_password(realm: k5test.K5Realm) -> None:
    ctx = krb5.init_context()
    princ = krb5.parse_name_flags(ctx, realm.user_princ.encode())
    creds_ctx = krb5.init_creds_init(ctx, princ)

    assert isinstance(creds_ctx, krb5.InitCredsContext)
    assert str(creds_ctx) == "InitCredsContext"

    krb5.init_creds_set_password(ctx, creds_ctx, realm.password("user").encode())
    krb5.init_creds_get(ctx, creds_ctx)

    creds = krb5.init_creds_get_creds(ctx, creds_ctx)
    assert isinstance(creds, krb5.Creds)
    assert str(creds) == "Creds"


def test_init_creds_set_password_invalid(realm: k5test.K5Realm) -> None:
    ctx = krb5.init_context()
    princ = krb5.parse_name_flags(ctx, realm.user_princ.encode())
    creds_ctx = krb5.init_creds_init(ctx, princ)

    assert isinstance(creds_ctx, krb5.InitCredsContext)
    assert str(creds_ctx) == "InitCredsContext"

    krb5.init_creds_set_password(ctx, creds_ctx, b"invalid")

    # Too many different error messages - just expect an error
    with pytest.raises(krb5.Krb5Error):
        krb5.init_creds_get(ctx, creds_ctx)



# Copyright: (c) 2021 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import k5test
import pytest

import krb5


def test_parse_principal(realm: k5test.K5Realm) -> None:
    name = f"role/abc\\/def\\@test@{realm.realm}"
    ctx = krb5.init_context()

    principal = krb5.parse_name_flags(ctx, name.encode())
    assert isinstance(principal, krb5.Principal)
    assert str(principal) == name
    assert repr(principal) == f"Principal({name})"
    assert principal.name == name.encode()
    assert isinstance(principal.addr, int)

    no_flags = krb5.parse_name_flags(ctx, name.encode(), flags=krb5.PrincipalParseFlags.none)
    assert isinstance(principal, krb5.Principal)
    assert str(principal) == str(no_flags)

    no_realm = krb5.parse_name_flags(ctx, b"role/abc\\/def\\@test", flags=krb5.PrincipalParseFlags.no_realm)
    assert isinstance(no_realm, krb5.Principal)
    if realm.provider == "mit":
        # MIT seems to have a bug here
        assert str(no_realm) == "role/abc\\/def@test"
    else:
        assert str(no_realm) == "role/abc\\/def\\@test"

    require_realm = krb5.parse_name_flags(ctx, name.encode(), flags=krb5.PrincipalParseFlags.require_realm)
    assert isinstance(require_realm, krb5.Principal)
    assert str(require_realm) == name

    enterprise = krb5.parse_name_flags(ctx, name.encode(), flags=krb5.PrincipalParseFlags.enterprise)
    assert isinstance(enterprise, krb5.Principal)
    assert str(enterprise) == f"role\\/abc\\/def\\@test\\@{realm.realm}@{realm.realm}"

    ignore_realm = krb5.parse_name_flags(ctx, name.encode(), flags=krb5.PrincipalParseFlags.ignore_realm)
    assert isinstance(ignore_realm, krb5.Principal)
    if realm.provider == "mit":
        assert str(ignore_realm) == "role/abc\\/def@test"
    else:
        assert str(ignore_realm) == "role/abc\\/def\\@test"


def test_parse_principal_no_realm_failure(realm: k5test.K5Realm) -> None:
    ctx = krb5.init_context()

    expected = "has realm present" if realm.provider == "mit" else "realm found in"
    with pytest.raises(krb5.Krb5Error, match=expected):
        krb5.parse_name_flags(ctx, realm.user_princ.encode(), flags=krb5.PrincipalParseFlags.no_realm)


def test_unparse_principal(realm: k5test.K5Realm) -> None:
    name = f"role/abc\\/def\\@test@{realm.realm}"
    ctx = krb5.init_context()
    principal = krb5.parse_name_flags(ctx, name.encode())

    normal = krb5.unparse_name_flags(ctx, principal)
    assert normal == b"role/abc\\/def\\@test@" + realm.realm.encode()

    no_flags = krb5.unparse_name_flags(ctx, principal, flags=krb5.PrincipalUnparseFlags.none)
    assert no_flags == normal

    no_realm = krb5.unparse_name_flags(ctx, principal, flags=krb5.PrincipalUnparseFlags.no_realm)
    if realm.provider == "mit":
        # MIT seems to have a bug here
        assert no_realm == b"role/abc\\/def@test"
    else:
        assert no_realm == b"role/abc\\/def\\@test"

    display = krb5.unparse_name_flags(ctx, principal, flags=krb5.PrincipalUnparseFlags.display)
    assert display == b"role/abc/def@test@" + realm.realm.encode()

    short = krb5.unparse_name_flags(ctx, principal, flags=krb5.PrincipalUnparseFlags.short)
    assert short == b"role/abc\\/def\\@test"

    krb5.set_default_realm(ctx, b"NEW.REALM")
    short = krb5.unparse_name_flags(ctx, principal, flags=krb5.PrincipalUnparseFlags.short)
    assert short == b"role/abc\\/def\\@test@" + realm.realm.encode()


@pytest.mark.requires_api("principal_get_realm")
def test_principal_get_realm() -> None:
    ctx = krb5.init_context()
    principal = krb5.parse_name_flags(ctx, b"username@REALM.COM")

    realm = krb5.principal_get_realm(ctx, principal)
    assert realm == b"REALM.COM"



# Copyright: (c) 2021 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import os

import k5test
import pytest

import krb5


def test_get_init_creds_opt_alloc() -> None:
    ctx = krb5.init_context()

    opt = krb5.get_init_creds_opt_alloc(ctx)
    assert isinstance(opt, krb5.GetInitCredsOpt)
    assert str(opt) == "GetInitCredsOpt"


def test_get_init_creds_opt_set_anonymous() -> None:
    ctx = krb5.init_context()
    opt = krb5.get_init_creds_opt_alloc(ctx)
    krb5.get_init_creds_opt_set_anonymous(opt, True)
    krb5.get_init_creds_opt_set_anonymous(opt, False)


def test_get_init_creds_opt_set_canonicalize() -> None:
    ctx = krb5.init_context()
    opt = krb5.get_init_creds_opt_alloc(ctx)
    krb5.get_init_creds_opt_set_canonicalize(opt, True)
    krb5.get_init_creds_opt_set_canonicalize(opt, False)


def test_get_init_creds_opt_set_etype_list() -> None:
    ctx = krb5.init_context()
    opt = krb5.get_init_creds_opt_alloc(ctx)
    krb5.get_init_creds_opt_set_etype_list(opt, [17, 18])
    krb5.get_init_creds_opt_set_etype_list(opt, [])


def test_get_init_creds_opt_set_forwardable() -> None:
    ctx = krb5.init_context()
    opt = krb5.get_init_creds_opt_alloc(ctx)
    krb5.get_init_creds_opt_set_forwardable(opt, True)
    krb5.get_init_creds_opt_set_forwardable(opt, False)


def test_get_init_creds_opt_set_proxiable() -> None:
    ctx = krb5.init_context()
    opt = krb5.get_init_creds_opt_alloc(ctx)
    krb5.get_init_creds_opt_set_proxiable(opt, True)
    krb5.get_init_creds_opt_set_proxiable(opt, False)


def test_get_init_creds_opt_set_renew_life() -> None:
    ctx = krb5.init_context()
    opt = krb5.get_init_creds_opt_alloc(ctx)
    krb5.get_init_creds_opt_set_renew_life(opt, 0)
    krb5.get_init_creds_opt_set_renew_life(opt, 1)
    krb5.get_init_creds_opt_set_renew_life(opt, 1024)


def test_get_init_creds_opt_set_salt() -> None:
    ctx = krb5.init_context()
    opt = krb5.get_init_creds_opt_alloc(ctx)
    krb5.get_init_creds_opt_set_salt(opt, b"\x00")


def test_get_init_creds_opt_set_salt_invalid_empty_string() -> None:
    ctx = krb5.init_context()
    opt = krb5.get_init_creds_opt_alloc(ctx)
    with pytest.raises(ValueError, match="salt cannot be an empty byte string"):
        krb5.get_init_creds_opt_set_salt(opt, b"")


def test_get_init_creds_opt_set_tkt_life() -> None:
    ctx = krb5.init_context()
    opt = krb5.get_init_creds_opt_alloc(ctx)
    krb5.get_init_creds_opt_set_tkt_life(opt, 0)
    krb5.get_init_creds_opt_set_tkt_life(opt, 1)
    krb5.get_init_creds_opt_set_tkt_life(opt, 1024)


@pytest.mark.requires_api("get_init_creds_opt_set_fast_flags")
def test_get_init_creds_opt_set_fast_flags() -> None:
    ctx = krb5.init_context()
    opt = krb5.get_init_creds_opt_alloc(ctx)
    krb5.get_init_creds_opt_set_fast_flags(ctx, opt, krb5.FastFlags.required)
    krb5.get_init_creds_opt_set_fast_flags(ctx, opt, krb5.FastFlags.none)


@pytest.mark.requires_api("get_init_creds_opt_set_fast_ccache")
def test_get_init_creds_opt_set_fast_ccache() -> None:
    ctx = krb5.init_context()
    opt = krb5.get_init_creds_opt_alloc(ctx)
    cc = krb5.cc_new_unique(ctx, b"MEMORY")
    krb5.get_init_creds_opt_set_fast_ccache(ctx, opt, cc)


@pytest.mark.requires_api("get_init_creds_opt_set_fast_ccache_name")
def test_get_init_creds_opt_set_fast_ccache_name() -> None:
    ctx = krb5.init_context()
    opt = krb5.get_init_creds_opt_alloc(ctx)
    cc = krb5.cc_new_unique(ctx, b"MEMORY")
    krb5.get_init_creds_opt_set_fast_ccache_name(ctx, opt, (cc.cache_type or b"") + b":" + (cc.name or b""))


@pytest.mark.requires_api("get_init_creds_opt_set_in_ccache")
def test_get_init_creds_opt_set_in_ccache() -> None:
    ctx = krb5.init_context()
    opt = krb5.get_init_creds_opt_alloc(ctx)
    cc = krb5.cc_new_unique(ctx, b"MEMORY")
    krb5.get_init_creds_opt_set_in_ccache(ctx, opt, cc)


@pytest.mark.requires_api("get_init_creds_opt_set_out_ccache")
def test_get_init_creds_opt_set_out_ccache() -> None:
    ctx = krb5.init_context()
    ccache = krb5.cc_new_unique(ctx, b"MEMORY")

    opt = krb5.get_init_creds_opt_alloc(ctx)
    krb5.get_init_creds_opt_set_out_ccache(ctx, opt, ccache)


@pytest.mark.requires_api("get_init_creds_opt_set_pac_request")
def test_get_init_creds_opt_set_pac_request() -> None:
    ctx = krb5.init_context()
    opt = krb5.get_init_creds_opt_alloc(ctx)
    krb5.get_init_creds_opt_set_pac_request(ctx, opt, True)
    krb5.get_init_creds_opt_set_pac_request(ctx, opt, False)


@pytest.mark.requires_api("get_init_creds_opt_set_default_flags")
def test_get_init_creds_opt_set_default_flags() -> None:
    ctx = krb5.init_context()
    opt = krb5.get_init_creds_opt_alloc(ctx)

    krb5.get_init_creds_opt_set_default_flags(ctx, opt)
    krb5.get_init_creds_opt_set_default_flags(ctx, opt, b"appname", b"realm")


@pytest.mark.requires_api("get_init_creds_opt_set_pa")
def test_get_init_creds_opt_set_pa() -> None:
    ctx = krb5.init_context()
    opt = krb5.get_init_creds_opt_alloc(ctx)

    krb5.get_init_creds_opt_set_pa(ctx, opt, b"attr", b"value")


@pytest.mark.requires_api("get_init_creds_opt_set_pa")
def test_get_init_creds_opt_set_pa_invalid_empty_value() -> None:
    ctx = krb5.init_context()
    opt = krb5.get_init_creds_opt_alloc(ctx)

    with pytest.raises(ValueError, match="attr must be set"):
        krb5.get_init_creds_opt_set_pa(ctx, opt, b"", b"value")

    with pytest.raises(ValueError, match="value must be set"):
        krb5.get_init_creds_opt_set_pa(ctx, opt, b"attr", b"")



# Copyright: (c) 2021 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import copy
import pathlib

import k5test
import pytest

import krb5


def test_kt_default(realm: k5test.K5Realm) -> None:
    ctx = krb5.init_context()

    kt = krb5.kt_default(ctx)
    assert isinstance(kt, krb5.KeyTab)
    assert isinstance(kt.addr, int)
    assert kt.name == realm.keytab.encode()
    assert kt.kt_type == b"FILE"

    assert repr(kt) == f"KeyTab(kt_type=FILE, name={realm.keytab})"
    assert str(kt) == f"FILE:{realm.keytab}"


def test_kt_default_name(realm: k5test.K5Realm) -> None:
    ctx = krb5.init_context()

    actual = krb5.kt_default_name(ctx)
    assert actual == realm.keytab.encode()


def test_kt_get_name(realm: k5test.K5Realm, tmp_path: pathlib.Path) -> None:
    ctx = krb5.init_context()

    kt = krb5.kt_resolve(ctx, f"FILE:{tmp_path / 'keytab'}".encode())
    actual = krb5.kt_get_name(ctx, kt)

    if realm.provider == "mit":
        assert actual == f"FILE:{tmp_path / 'keytab'}".encode()
    else:
        assert actual == f"{tmp_path / 'keytab'}".encode()


def test_kt_get_full_name(tmp_path: pathlib.Path) -> None:
    ctx = krb5.init_context()

    kt = krb5.kt_resolve(ctx, f"FILE:{tmp_path / 'keytab'}".encode())
    assert krb5.kt_get_full_name(ctx, kt) == f"FILE:{tmp_path / 'keytab'}".encode()


def test_kt_get_type(tmp_path: pathlib.Path) -> None:
    ctx = krb5.init_context()

    kt = krb5.kt_resolve(ctx, f"FILE:{tmp_path / 'keytab'}".encode())
    assert krb5.kt_get_type(ctx, kt) == b"FILE"


def test_kt_enumerate(realm: k5test.K5Realm, tmp_path: pathlib.Path) -> None:
    ctx = krb5.init_context()
    kt = krb5.kt_resolve(ctx, f"FILE:{tmp_path / 'keytab'}".encode())

    msg_pattern = "Key table file '.*' not found" if realm.provider == "mit" else "No such file or directory"
    with pytest.raises(krb5.Krb5Error, match=msg_pattern):
        list(kt)

    princ = krb5.parse_name_flags(ctx, b"user@DOMAIN.COM")
    key_block = krb5.init_keyblock(ctx, 17, b"\x00" * 16)
    krb5.kt_add_entry(ctx, kt, princ, 1, 0, key_block)

    entries = list(kt)
    assert len(entries) == 1
    assert entries[0].principal.name == b"user@DOMAIN.COM"
    assert entries[0].kvno == 1
    assert isinstance(entries[0].timestamp, int)
    copied_princ = copy.copy(entries[0].principal)

    krb5.kt_remove_entry(ctx, kt, entries[0])
    del entries[0]

    entries = list(kt)
    assert entries == []

    # Tests that the copied principal outlives the kt/entry
    del kt
    assert copied_princ.name == b"user@DOMAIN.COM"


def test_kt_get_entry_empty(realm: k5test.K5Realm, tmp_path: pathlib.Path) -> None:
    ctx = krb5.init_context()
    kt = krb5.kt_resolve(ctx, f"FILE:{tmp_path / 'keytab'}".encode())
    princ = krb5.parse_name_flags(ctx, b"user@DOMAIN.COM")

    msg_pattern = "Key table file '.*' not found" if realm.provider == "mit" else "No such file or directory"
    with pytest.raises(krb5.Krb5Error, match=msg_pattern):
        krb5.kt_get_entry(ctx, kt, princ)


def test_kt_get_entry(realm: k5test.K5Realm, tmp_path: pathlib.Path) -> None:
    ctx = krb5.init_context()
    kt = krb5.kt_resolve(ctx, f"FILE:{tmp_path / 'keytab'}".encode())
    princ = krb5.parse_name_flags(ctx, b"user@DOMAIN.COM")
    key_block = krb5.init_keyblock(ctx, 17, b"\xff" * 16)
    assert key_block.enctype == 17
    assert key_block.data == b"\xff" * 16

    krb5.kt_add_entry(ctx, kt, princ, 1, 0, key_block)

    entry = krb5.kt_get_entry(ctx, kt, princ)
    assert entry.principal.name == princ.name
    assert entry.kvno == 1
    assert entry.key.enctype == 17
    assert entry.key.data == key_block.data

    krb5.kt_remove_entry(ctx, kt, entry)
    assert list(kt) == []

    msg_pattern = "No key table entry found" if realm.provider == "mit" else "Failed to find .* in keytab .*"
    with pytest.raises(krb5.Krb5Error, match=msg_pattern):
        krb5.kt_get_entry(ctx, kt, princ)


def test_kt_get_entry_multiple_kvno(realm: k5test.K5Realm, tmp_path: pathlib.Path) -> None:
    ctx = krb5.init_context()
    kt = krb5.kt_resolve(ctx, f"FILE:{tmp_path / 'keytab'}".encode())
    princ = krb5.parse_name_flags(ctx, b"user@DOMAIN.COM")
    key_block = krb5.init_keyblock(ctx, 17, b"\x00" * 16)
    krb5.kt_add_entry(ctx, kt, princ, 1, 0, key_block)
    krb5.kt_add_entry(ctx, kt, princ, 2, 0, key_block)

    entry = krb5.kt_get_entry(ctx, kt, princ)
    assert entry.principal.name == princ.name
    assert entry.kvno == 2

    entry = krb5.kt_get_entry(ctx, kt, princ, kvno=1)
    assert entry.principal.name == princ.name
    assert entry.kvno == 1

    msg_pattern = (
        "Key version number for principal in key table is incorrect"
        if realm.provider == "mit"
        else "Failed to find .* in keytab .*"
    )
    with pytest.raises(krb5.Krb5Error, match=msg_pattern):
        krb5.kt_get_entry(ctx, kt, princ, kvno=3)


def test_kt_get_entry_multiple_enctype(realm: k5test.K5Realm, tmp_path: pathlib.Path) -> None:
    ctx = krb5.init_context()
    kt = krb5.kt_resolve(ctx, f"FILE:{tmp_path / 'keytab'}".encode())
    princ = krb5.parse_name_flags(ctx, b"user@DOMAIN.COM")
    krb5.kt_add_entry(ctx, kt, princ, 1, 0, krb5.init_keyblock(ctx, 17, b"\x00" * 16))
    krb5.kt_add_entry(ctx, kt, princ, 2, 0, krb5.init_keyblock(ctx, 18, b"\x00" * 32))

    entry = krb5.kt_get_entry(ctx, kt, princ)
    assert entry.principal.name == princ.name
    assert entry.kvno == 2
    assert entry.key.enctype == 18

    entry = krb5.kt_get_entry(ctx, kt, princ, enctype=17)
    assert entry.principal.name == princ.name
    assert entry.kvno == 1
    assert entry.key.enctype == 17

    msg_pattern = "No key table entry found" if realm.provider == "mit" else "Failed to find .* in keytab .*"
    with pytest.raises(krb5.Krb5Error, match=msg_pattern):
        krb5.kt_get_entry(ctx, kt, princ, enctype=16)


def test_kt_read_service_key_empty(realm: k5test.K5Realm, tmp_path: pathlib.Path) -> None:
    ctx = krb5.init_context()
    kt = krb5.kt_resolve(ctx, f"FILE:{tmp_path / 'keytab'}".encode())
    princ = krb5.parse_name_flags(ctx, b"user@DOMAIN.COM")

    msg_pattern = "Key table file '.*' not found" if realm.provider == "mit" else "No such file or directory"
    with pytest.raises(krb5.Krb5Error, match=msg_pattern):
        krb5.kt_read_service_key(ctx, kt.name, princ)


def test_kt_read_service_key(tmp_path: pathlib.Path) -> None:
    ctx = krb5.init_context()
    kt = krb5.kt_resolve(ctx, f"FILE:{tmp_path / 'keytab'}".encode())
    princ = krb5.parse_name_flags(ctx, b"user@DOMAIN.COM")
    key_block = krb5.init_keyblock(ctx, 17, b"\xff" * 16)
    assert key_block.enctype == 17
    assert key_block.data == b"\xff" * 16

    krb5.kt_add_entry(ctx, kt, princ, 1, 0, key_block)

    key = krb5.kt_read_service_key(ctx, kt.name, princ)
    assert key.enctype == 17
    assert key.data == b"\xff" * 16


def test_kt_read_service_key_multiple_kvno(realm: k5test.K5Realm, tmp_path: pathlib.Path) -> None:
    ctx = krb5.init_context()
    kt = krb5.kt_resolve(ctx, f"FILE:{tmp_path / 'keytab'}".encode())
    princ = krb5.parse_name_flags(ctx, b"user@DOMAIN.COM")
    krb5.kt_add_entry(ctx, kt, princ, 1, 0, krb5.init_keyblock(ctx, 17, b"\x00" * 16))
    krb5.kt_add_entry(ctx, kt, princ, 2, 0, krb5.init_keyblock(ctx, 18, b"\x11" * 32))

    key = krb5.kt_read_service_key(ctx, kt.name, princ)
    assert key.enctype == 18
    assert key.data == b"\x11" * 32

    key = krb5.kt_read_service_key(ctx, kt.name, princ, kvno=1)
    assert key.enctype == 17
    assert key.data == b"\x00" * 16

    msg_pattern = (
        "Key version number for principal in key table is incorrect"
        if realm.provider == "mit"
        else "Failed to find .* in keytab .*"
    )
    with pytest.raises(krb5.Krb5Error, match=msg_pattern):
        krb5.kt_read_service_key(ctx, kt.name, princ, kvno=3)


def test_kt_read_service_key_multiple_enctype(realm: k5test.K5Realm, tmp_path: pathlib.Path) -> None:
    ctx = krb5.init_context()
    kt = krb5.kt_resolve(ctx, f"FILE:{tmp_path / 'keytab'}".encode())
    princ = krb5.parse_name_flags(ctx, b"user@DOMAIN.COM")
    krb5.kt_add_entry(ctx, kt, princ, 1, 0, krb5.init_keyblock(ctx, 17, b"\x00" * 16))
    krb5.kt_add_entry(ctx, kt, princ, 2, 0, krb5.init_keyblock(ctx, 18, b"\x11" * 32))

    key = krb5.kt_read_service_key(ctx, kt.name, princ)
    assert key.enctype == 18
    assert key.data == b"\x11" * 32

    key = krb5.kt_read_service_key(ctx, kt.name, princ, enctype=17)
    assert key.enctype == 17
    assert key.data == b"\x00" * 16

    msg_pattern = "No key table entry found" if realm.provider == "mit" else "Failed to find .* in keytab .*"
    with pytest.raises(krb5.Krb5Error, match=msg_pattern):
        krb5.kt_read_service_key(ctx, kt.name, princ, enctype=16)


def test_kt_resolve(tmp_path: pathlib.Path) -> None:
    ctx = krb5.init_context()

    kt = krb5.kt_resolve(ctx, f"FILE:{tmp_path / 'keytab'}".encode())
    assert isinstance(kt, krb5.KeyTab)
    assert kt.kt_type == b"FILE"
    assert kt.name == f"{tmp_path / 'keytab'}".encode()
    assert str(kt) == f"FILE:{tmp_path / 'keytab'}"
    assert repr(kt) == f"KeyTab(kt_type=FILE, name={tmp_path / 'keytab'})"


@pytest.mark.requires_api("kt_client_default")
def test_kt_client_default(realm: k5test.K5Realm) -> None:
    ctx = krb5.init_context()

    kt = krb5.kt_client_default(ctx)
    assert isinstance(kt, krb5.KeyTab)
    assert isinstance(kt.addr, int)
    assert kt.name == realm.client_keytab.encode()
    assert kt.kt_type == b"FILE"

    assert repr(kt) == f"KeyTab(kt_type=FILE, name={realm.client_keytab})"
    assert str(kt) == f"FILE:{realm.client_keytab}"


@pytest.mark.requires_api("kt_dup")
def test_kt_dup(tmp_path: pathlib.Path) -> None:
    ctx = krb5.init_context()

    kt = krb5.kt_resolve(ctx, f"FILE:{tmp_path / 'keytab'}".encode())
    kt_name = kt.name
    kt_addr = kt.addr

    copied_kt = krb5.kt_dup(ctx, kt)
    del kt
    assert isinstance(copied_kt, krb5.KeyTab)
    assert copied_kt.kt_type == b"FILE"
    assert copied_kt.name == kt_name
    assert copied_kt.addr != kt_addr


@pytest.mark.requires_api("kt_have_content")
def test_kt_have_content(tmp_path: pathlib.Path) -> None:
    ctx = krb5.init_context()

    kt = krb5.kt_resolve(ctx, f"FILE:{tmp_path / 'keytab'}".encode())

    assert krb5.kt_have_content(ctx, kt) is False

    princ = krb5.parse_name_flags(ctx, b"user@DOMAIN.COM")
    key_block = krb5.init_keyblock(ctx, 17, b"\x00" * 16)
    krb5.kt_add_entry(ctx, kt, princ, 1, 0, key_block)

    assert krb5.kt_have_content(ctx, kt)

    entry = list(kt)[0]
    krb5.kt_remove_entry(ctx, kt, entry)

    assert krb5.kt_have_content(ctx, kt) is False



# Copyright: (c) 2021 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import enum
import typing

from krb5._context import Context

class PrincipalParseFlags(enum.IntEnum):
    """Flags used to control :meth:`parse_name_flags`."""

    none: PrincipalParseFlags = ...  #: No parse flags set
    no_realm: PrincipalParseFlags = ...  #: Error if realm is present
    require_realm: PrincipalParseFlags = ...  #: Error if realm is not present
    enterprise: PrincipalParseFlags = ...  #: Create single-component enterprise principal
    ignore_realm: PrincipalParseFlags = ...  #: Ignore realm if present

class PrincipalUnparseFlags(enum.IntEnum):
    """Flags used to control :meth:`unparse_name_flags`."""

    none: PrincipalUnparseFlags = ...  #: No unparse flags set
    short: PrincipalUnparseFlags = ...  #: Omit realm if it is the local realm
    no_realm: PrincipalUnparseFlags = ...  #: Omit realm always
    display: PrincipalUnparseFlags = ...  #: Don't escape special characters

class Principal:
    """Kerberos Principal object.

    This class represents a Kerberos principal.

    Args:
        context: Krb5 context.
    """

    def __copy__(self) -> "Principal":
        """Create a copy of the principal object."""
    @property
    def addr(self) -> typing.Optional[int]:
        """The raw krb5_principal pointer address of this credential cache."""
    @property
    def name(self) -> typing.Optional[bytes]:
        """The name of the principal."""

def copy_principal(
    context: Context,
    principal: Principal,
) -> Principal:
    """Copy a principal.

    Creates a copy of the principal specified.

    Args:
        context: Krb5 context.
        principal: The principal to copy.

    Returns:
        Principal: The copy of the principal.
    """

def parse_name_flags(
    context: Context,
    name: bytes,
    flags: typing.Union[int, PrincipalParseFlags] = PrincipalParseFlags.none,
) -> Principal:
    """Create a Kerberos principal.

    Convert a string principal name to a Kerberos principal object.

    Args:
        context: Krb5 context.
        name: The principal name to parse.
        flags: Optional flags to control how the string is parsed.

    Returns:
        Principal: The Kerberos principal parsed from the string.
    """

def unparse_name_flags(
    context: Context,
    principal: Principal,
    flags: typing.Union[int, PrincipalUnparseFlags] = PrincipalUnparseFlags.none,
) -> bytes:
    """Get the Kerberos principal name.

    Converts a Kerberos principal to a string representation.

    args:
        context: Krb5 context.
        principal: The principal to convert from.
        flags: Optional flags to control how the string is generated.

    Returns:
        bytes: The principal as a byte string.
    """
